@@ -37,19 +37,33 @@
 /* Timeout before the fade-out. */
 #define TIMEOUT_FADE_OUT 250
 
-/* Number of tolerance pixels. */
-#define TOLERANCE_PIXELS 3
+/* Thumb radius in pixels (higher values are automatically clamped). */
+#define THUMB_RADIUS 3
+
+/* Number of tolerance pixels, before hiding the thumb. */
+#define TOLERANCE_FADE 3
+
+#ifndef USE_GTK3
+typedef struct {
+  gdouble red;
+  gdouble green;
+  gdouble blue;
+  gdouble alpha;
+} GdkRGBA;
+#endif
 
 struct _OsThumbPrivate {
   GtkOrientation orientation;
   GtkWidget *grabbed_widget;
   OsAnimation *animation;
-  gboolean button_press_event;
-  gboolean motion_notify_event;
+  OsEvent event;
   gboolean can_rgba;
+  gboolean detached;
   gboolean use_tolerance;
   gint pointer_x;
   gint pointer_y;
+  gint pointer_x_root;
+  gint pointer_y_root;
   guint32 source_id;
 };
 
@@ -59,12 +73,14 @@ enum {
   LAST_ARG
 };
 
-static void os_thumb_fade_out_cb (gfloat weight, gpointer user_data);
-static gboolean os_thumb_timeout_fade_out_cb (gpointer user_data);
 static gboolean os_thumb_button_press_event (GtkWidget *widget, GdkEventButton *event);
 static gboolean os_thumb_button_release_event (GtkWidget *widget, GdkEventButton *event);
 static void os_thumb_composited_changed (GtkWidget *widget);
+#ifdef USE_GTK3
+static gboolean os_thumb_draw (GtkWidget *widget, cairo_t *cr);
+#else
 static gboolean os_thumb_expose (GtkWidget *widget, GdkEventExpose *event);
+#endif
 static gboolean os_thumb_leave_notify_event (GtkWidget *widget, GdkEventCrossing *event);
 static gboolean os_thumb_motion_notify_event (GtkWidget *widget, GdkEventMotion *event);
 static void os_thumb_map (GtkWidget *widget);
@@ -77,52 +93,37 @@ static void os_thumb_finalize (GObject *object);
 static void os_thumb_get_property (GObject *object, guint prop_id, GValue *value, GParamSpec *pspec);
 static void os_thumb_set_property (GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec);
 
-/* Private functions. */
+/* Private functions */
 
-/* Draw a rounded rectangle. */
+/* callback called by the fade-out animation */
 static void
-os_cairo_draw_rounded_rect (cairo_t *cr,
-                            gdouble  x,
-                            gdouble  y,
-                            gdouble  width,
-                            gdouble  height,
-                            gdouble  radius)
-{
-  if (radius < 1)
-    {
-      cairo_rectangle (cr, x, y, width, height);
-      return;
-    }
-
-  radius = MIN (radius, MIN (width / 2.0, height / 2.0));
-
-  cairo_move_to (cr, x + radius, y);
-
-  cairo_arc (cr, x + width - radius, y + radius, radius, G_PI * 1.5, G_PI * 2);
-  cairo_arc (cr, x + width - radius, y + height - radius, radius, 0, G_PI * 0.5);
-  cairo_arc (cr, x + radius, y + height - radius, radius, G_PI * 0.5, G_PI);
-  cairo_arc (cr, x + radius, y + radius, radius, G_PI, G_PI * 1.5);
-}
-
-static void
-os_thumb_fade_out_cb (gfloat weight,
-                      gpointer user_data)
+fade_out_cb (gfloat   weight,
+             gpointer user_data)
 {
   OsThumb *thumb;
-  OsThumbPrivate *priv;
 
   thumb = OS_THUMB (user_data);
 
-  priv = thumb->priv;
-
   if (weight < 1.0f)
     gtk_window_set_opacity (GTK_WINDOW (thumb), fabs (weight - 1.0f));
   else
     gtk_widget_hide (GTK_WIDGET (thumb));
 }
 
+/* stop_func called by the fade-out animation */
+static void
+fade_out_stop_cb (gpointer user_data)
+{
+  OsThumb *thumb;
+
+  thumb = OS_THUMB (user_data);
+
+  gtk_window_set_opacity (GTK_WINDOW (thumb), 1.0f);
+}
+
+/* timeout before starting the fade-out animation */
 static gboolean
-os_thumb_timeout_fade_out_cb (gpointer user_data)
+timeout_fade_out_cb (gpointer user_data)
 {
   OsThumb *thumb;
   OsThumbPrivate *priv;
@@ -137,8 +138,6 @@ os_thumb_timeout_fade_out_cb (gpointer user_data)
   return FALSE;
 }
 
-/* Type definition. */
-
 G_DEFINE_TYPE (OsThumb, os_thumb, GTK_TYPE_WINDOW);
 
 static void
@@ -153,7 +152,11 @@ os_thumb_class_init (OsThumbClass *class)
   widget_class->button_press_event   = os_thumb_button_press_event;
   widget_class->button_release_event = os_thumb_button_release_event;
   widget_class->composited_changed   = os_thumb_composited_changed;
+#ifdef USE_GTK3
+  widget_class->draw                 = os_thumb_draw;
+#else
   widget_class->expose_event         = os_thumb_expose;
+#endif
   widget_class->leave_notify_event   = os_thumb_leave_notify_event;
   widget_class->map                  = os_thumb_map;
   widget_class->motion_notify_event  = os_thumb_motion_notify_event;
@@ -188,15 +191,18 @@ os_thumb_init (OsThumb *thumb)
                                              OsThumbPrivate);
   priv = thumb->priv;
 
+  priv->event = OS_EVENT_NONE;
+
   priv->can_rgba = FALSE;
+  priv->detached = FALSE;
+  priv->use_tolerance = FALSE;
 
   priv->source_id = 0;
   priv->animation = os_animation_new (RATE_FADE_OUT, DURATION_FADE_OUT,
-                                      os_thumb_fade_out_cb, NULL, thumb);
+                                      fade_out_cb, NULL, thumb);
 
   gtk_window_set_skip_pager_hint (GTK_WINDOW (thumb), TRUE);
   gtk_window_set_skip_taskbar_hint (GTK_WINDOW (thumb), TRUE);
-  /* gtk_window_set_has_resize_grip (GTK_WINDOW (thumb), FALSE); */
   gtk_window_set_decorated (GTK_WINDOW (thumb), FALSE);
   gtk_window_set_focus_on_map (GTK_WINDOW (thumb), FALSE);
   gtk_window_set_accept_focus (GTK_WINDOW (thumb), FALSE);
@@ -210,43 +216,6 @@ os_thumb_init (OsThumb *thumb)
   os_thumb_composited_changed (GTK_WIDGET (thumb));
 }
 
-static void
-os_thumb_dispose (GObject *object)
-{
-  OsThumb *thumb;
-  OsThumbPrivate *priv;
-
-  thumb = OS_THUMB (object);
-  priv = thumb->priv;
-
-  if (priv->source_id != 0)
-    {
-      g_source_remove (priv->source_id);
-      priv->source_id = 0;
-    }
-
-  if (priv->animation != NULL)
-    {
-      g_object_unref (priv->animation);
-      priv->animation = NULL;
-    }
-
-  if (priv->grabbed_widget != NULL)
-    {
-      g_object_unref (priv->grabbed_widget);
-      priv->grabbed_widget = NULL;
-    }
-
-  G_OBJECT_CLASS (os_thumb_parent_class)->dispose (object);
-}
-
-static void
-os_thumb_finalize (GObject *object)
-{
-  G_OBJECT_CLASS (os_thumb_parent_class)->finalize (object);
-}
-
-
 static gboolean
 os_thumb_button_press_event (GtkWidget      *widget,
                              GdkEventButton *event)
@@ -265,8 +234,7 @@ os_thumb_button_press_event (GtkWidget      *widget,
 
   /* Stop the animation on user interaction,
    * the button_press_event. */
-  os_animation_stop (priv->animation);
-  gtk_window_set_opacity (GTK_WINDOW (widget), 1.0f);
+  os_animation_stop (priv->animation, fade_out_stop_cb);
 
   if (event->type == GDK_BUTTON_PRESS)
     {
@@ -276,9 +244,11 @@ os_thumb_button_press_event (GtkWidget      *widget,
 
           priv->pointer_x = event->x;
           priv->pointer_y = event->y;
+          priv->pointer_x_root = event->x_root;
+          priv->pointer_y_root = event->y_root;
 
-          priv->button_press_event = TRUE;
-          priv->motion_notify_event = FALSE;
+          priv->event |= OS_EVENT_BUTTON_PRESS;
+          priv->event &= ~(OS_EVENT_MOTION_NOTIFY);
 
           priv->use_tolerance = TRUE;
 
@@ -308,8 +278,7 @@ os_thumb_button_release_event (GtkWidget      *widget,
         {
           gtk_grab_remove (widget);
 
-          priv->button_press_event = FALSE;
-          priv->motion_notify_event = FALSE;
+          priv->event &= ~(OS_EVENT_BUTTON_PRESS | OS_EVENT_MOTION_NOTIFY);
 
           gtk_widget_queue_draw (widget);
         }
@@ -332,49 +301,368 @@ os_thumb_composited_changed (GtkWidget *widget)
   if (gdk_screen_is_composited (gtk_widget_get_screen (widget)))
     {
       GdkVisual *visual;
+      guint32 red_mask;
+      guint32 green_mask;
+      guint32 blue_mask;
 
       visual = gtk_widget_get_visual (widget);
 
-      if (visual->depth == 32 && (visual->red_mask   == 0xff0000 &&
-                                  visual->green_mask == 0x00ff00 &&
-                                  visual->blue_mask  == 0x0000ff))
+      gdk_visual_get_red_pixel_details (visual, &red_mask, NULL, NULL);
+      gdk_visual_get_green_pixel_details (visual, &green_mask, NULL, NULL);
+      gdk_visual_get_blue_pixel_details (visual, &blue_mask, NULL, NULL);
+
+      if (gdk_visual_get_depth (visual) == 32 && (red_mask   == 0xff0000 &&
+                                                  green_mask == 0x00ff00 &&
+                                                  blue_mask  == 0x0000ff))   
         priv->can_rgba = TRUE;
     }
 
   gtk_widget_queue_draw (widget);
 }
 
+/* simplified wrapper of cairo_pattern_add_color_stop_rgba */
+static void
+pattern_add_gdk_rgba_stop (cairo_pattern_t *pat,
+                           gdouble          stop,
+                           const GdkRGBA   *color,
+                           gdouble          alpha)
+{
+  cairo_pattern_add_color_stop_rgba (pat, stop, color->red, color->green, color->blue, alpha);
+}
+
+/* simplified wrapper of cairo_set_source_rgba */
+static void
+set_source_gdk_rgba (cairo_t       *cr,
+                     const GdkRGBA *color,
+                     gdouble        alpha)
+{
+  cairo_set_source_rgba (cr, color->red, color->green, color->blue, alpha);
+}
+
+/* draw an arrow using cairo */
+static void
+draw_arrow (cairo_t       *cr,
+            const GdkRGBA *color,
+            gdouble        x,
+            gdouble        y,
+            gdouble        width,
+            gdouble        height)
+{
+  cairo_save (cr);
+
+  cairo_translate (cr, x, y);
+  cairo_move_to (cr, -width / 2, -height / 2);
+  cairo_line_to (cr, 0, height / 2);
+  cairo_line_to (cr, width / 2, -height / 2);
+  cairo_close_path (cr);
+
+  set_source_gdk_rgba (cr, color, 0.75);
+  cairo_fill_preserve (cr);
+
+  set_source_gdk_rgba (cr, color, 1.0);
+  cairo_stroke (cr);
+
+  cairo_restore (cr);
+}
+
+/* draw a rounded rectangle using cairo */
+static void
+draw_round_rect (cairo_t *cr,
+                 gdouble  x,
+                 gdouble  y,
+                 gdouble  width,
+                 gdouble  height,
+                 gdouble  radius)
+{
+  radius = MIN (radius, MIN (width / 2.0, height / 2.0));
+
+  if (radius < 1)
+    {
+      cairo_rectangle (cr, x, y, width, height);
+      return;
+    }
+
+  cairo_move_to (cr, x + radius, y);
+
+  cairo_arc (cr, x + width - radius, y + radius, radius, G_PI * 1.5, G_PI * 2);
+  cairo_arc (cr, x + width - radius, y + height - radius, radius, 0, G_PI * 0.5);
+  cairo_arc (cr, x + radius, y + height - radius, radius, G_PI * 0.5, G_PI);
+  cairo_arc (cr, x + radius, y + radius, radius, G_PI, G_PI * 1.5);
+}
+
+/* convert rgb to hls */
+static void
+rgb_to_hls (gdouble *r,
+            gdouble *g,
+            gdouble *b)
+{
+  gdouble min;
+  gdouble max;
+  gdouble red;
+  gdouble green;
+  gdouble blue;
+  gdouble h, l, s;
+  gdouble delta;
+
+  h = 0;
+
+  red = *r;
+  green = *g;
+  blue = *b;
+
+  if (red > green)
+    {
+      if (red > blue)
+        max = red;
+      else
+        max = blue;
+
+      if (green < blue)
+        min = green;
+      else
+        min = blue;
+    }
+  else
+    {
+      if (green > blue)
+        max = green;
+      else
+        max = blue;
+
+      if (red < blue)
+        min = red;
+      else
+        min = blue;
+    }
+
+  l = (max + min) / 2;
+  if (fabs (max - min) < 0.0001)
+    {
+      h = 0;
+      s = 0;
+    }
+  else
+    {
+      if (l <= 0.5)
+        s = (max - min) / (max + min);
+      else
+        s = (max - min) / (2 - max - min);
+
+      delta = max - min;
+      if (red == max)
+        h = (green - blue) / delta;
+      else if (green == max)
+        h = 2 + (blue - red) / delta;
+      else if (blue == max)
+        h = 4 + (red - green) / delta;
+
+      h *= 60;
+      if (h < 0.0)
+        h += 360;
+    }
+
+  *r = h;
+  *g = l;
+  *b = s;
+}
+
+/* convert hls to rgb */
+static void
+hls_to_rgb (gdouble *h,
+            gdouble *l,
+            gdouble *s)
+{
+  gdouble hue;
+  gdouble lightness;
+  gdouble saturation;
+  gdouble m1, m2;
+  gdouble r, g, b;
+
+  lightness = *l;
+  saturation = *s;
+
+  if (lightness <= 0.5)
+    m2 = lightness * (1 + saturation);
+  else
+    m2 = lightness + saturation - lightness * saturation;
+
+  m1 = 2 * lightness - m2;
+
+  if (saturation == 0)
+    {
+      *h = lightness;
+      *l = lightness;
+      *s = lightness;
+    }
+  else
+    {
+      hue = *h + 120;
+      while (hue > 360)
+        hue -= 360;
+      while (hue < 0)
+        hue += 360;
+
+      if (hue < 60)
+        r = m1 + (m2 - m1) * hue / 60;
+      else if (hue < 180)
+        r = m2;
+      else if (hue < 240)
+        r = m1 + (m2 - m1) * (240 - hue) / 60;
+      else
+        r = m1;
+
+      hue = *h;
+      while (hue > 360)
+        hue -= 360;
+      while (hue < 0)
+        hue += 360;
+
+      if (hue < 60)
+        g = m1 + (m2 - m1) * hue / 60;
+      else if (hue < 180)
+        g = m2;
+      else if (hue < 240)
+        g = m1 + (m2 - m1) * (240 - hue) / 60;
+      else
+        g = m1;
+
+      hue = *h-120;
+      while (hue > 360)
+        hue -= 360;
+      while (hue < 0)
+        hue += 360;
+
+      if (hue < 60)
+        b = m1 + (m2 - m1) * hue / 60;
+      else if (hue < 180)
+        b = m2;
+      else if (hue < 240)
+        b = m1 + (m2 - m1) * (240 - hue) / 60;
+      else
+        b = m1;
+
+      *h = r;
+      *l = g;
+      *s = b;
+    }
+}
+
+/* shade a GdkRGBA color */
+static void
+shade_gdk_rgba (const GdkRGBA *a,
+                gfloat         k,
+                GdkRGBA       *b)
+{
+  gdouble red;
+  gdouble green;
+  gdouble blue;
+
+  red   = a->red;
+  green = a->green;
+  blue  = a->blue;
+
+  if (k == 1.0)
+    {
+      b->red = red;
+      b->green = green;
+      b->blue = blue;
+      return;
+    }
+
+  rgb_to_hls (&red, &green, &blue);
+
+  green *= k;
+  if (green > 1.0)
+    green = 1.0;
+  else if (green < 0.0)
+    green = 0.0;
+
+  blue *= k;
+  if (blue > 1.0)
+    blue = 1.0;
+  else if (blue < 0.0)
+    blue = 0.0;
+
+  hls_to_rgb (&red, &green, &blue);
+
+  b->red = red;
+  b->green = green;
+  b->blue = blue;
+  b->alpha = a->alpha;
+}
+
+#ifndef USE_GTK3
+/* convert a GdkColor to GdkRGBA */
+static void
+convert_gdk_color_to_gdk_rgba (GdkColor *color,
+                               GdkRGBA  *rgba)
+{  
+  rgba->red = (gdouble) color->red / (gdouble) 65535;
+  rgba->green = (gdouble) color->green / (gdouble) 65535;
+  rgba->blue = (gdouble) color->blue / (gdouble) 65535;
+
+  rgba->alpha = 1.0;
+}
+#endif
+
 static gboolean
+#ifdef USE_GTK3
+os_thumb_draw (GtkWidget *widget,
+               cairo_t   *cr)
+{
+  GtkStyleContext *style_context;
+#else
 os_thumb_expose (GtkWidget      *widget,
                  GdkEventExpose *event)
 {
   GtkAllocation allocation;
-  GtkStateType state_type_down, state_type_up;
+  cairo_t *cr;
+  GtkStyle *style;
+#endif
+  GdkRGBA bg, bg_active, bg_selected;
+  GdkRGBA bg_arrow_up, bg_arrow_down;
+  GdkRGBA bg_shadow, bg_dark_line, bg_bright_line;
+  GdkRGBA arrow_color;
   OsThumb *thumb;
   OsThumbPrivate *priv;
   cairo_pattern_t *pat;
-  cairo_t *cr;
-  gint x, y, width, height;
+  gint width, height;
   gint radius;
 
   thumb = OS_THUMB (widget);
   priv = thumb->priv;
 
-  state_type_down = GTK_STATE_NORMAL;
-  state_type_up = GTK_STATE_NORMAL;
+  radius = priv->can_rgba ? THUMB_RADIUS : 0;
 
+#ifdef USE_GTK3
+  width = gtk_widget_get_allocated_width (widget);
+  height = gtk_widget_get_allocated_height (widget);
+
+  style_context = gtk_widget_get_style_context (widget);
+
+  gtk_style_context_get_background_color (style_context, gtk_widget_get_state_flags (widget), &bg);
+  gtk_style_context_get_background_color (style_context, GTK_STATE_FLAG_ACTIVE, &bg_active);
+  gtk_style_context_get_background_color (style_context, GTK_STATE_FLAG_SELECTED, &bg_selected);
+  gtk_style_context_get_color (style_context, gtk_widget_get_state_flags (widget), &arrow_color);
+#else
   gtk_widget_get_allocation (widget, &allocation);
-
-  x = 0;
-  y = 0;
+ 
   width = allocation.width;
   height = allocation.height;
-  radius = priv->can_rgba ? 18 : 0;
+
+  style = gtk_widget_get_style (widget);
+
+  convert_gdk_color_to_gdk_rgba (&style->bg[gtk_widget_get_state (widget)], &bg);
+  convert_gdk_color_to_gdk_rgba (&style->bg[GTK_STATE_ACTIVE], &bg_active);
+  convert_gdk_color_to_gdk_rgba (&style->bg[GTK_STATE_SELECTED], &bg_selected);
+  convert_gdk_color_to_gdk_rgba (&style->fg[gtk_widget_get_state (widget)], &arrow_color);
 
   cr = gdk_cairo_create (gtk_widget_get_window (widget));
+#endif
 
-  cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
-  cairo_set_source_rgba (cr, 0, 0, 0, 0);
+  cairo_save (cr);
+
+  cairo_set_operator (cr, CAIRO_OPERATOR_CLEAR);
   cairo_paint (cr);
 
   cairo_save (cr);
@@ -382,171 +670,158 @@ os_thumb_expose (GtkWidget      *widget,
   width--;
   height--;
 
-  cairo_set_line_width (cr, 1);
+  cairo_set_line_width (cr, 1.0);
   cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
 
-  os_cairo_draw_rounded_rect (cr, x, y, width, height, radius);
+  /* background */
+  draw_round_rect (cr, 0, 0, width, height, radius);
+
+  set_source_gdk_rgba (cr, &bg, 1.0);
+  cairo_fill_preserve (cr);
+
+  /* background pattern from top to bottom */
+  shade_gdk_rgba (&bg, 1.3, &bg_arrow_up);
+  shade_gdk_rgba (&bg, 0.7, &bg_arrow_down);
 
   if (priv->orientation == GTK_ORIENTATION_VERTICAL)
-    pat = cairo_pattern_create_linear (x, y, width + x, y);
+    pat = cairo_pattern_create_linear (0, 0, 0, height);
   else
-    pat = cairo_pattern_create_linear (x, y, x, height + y);
+    pat = cairo_pattern_create_linear (0, 0, width, 0);
 
-  cairo_pattern_add_color_stop_rgba (pat, 0.0, 0.95, 0.95, 0.95, 1.0);
-  cairo_pattern_add_color_stop_rgba (pat, 1.0, 0.8, 0.8, 0.8, 1.0);
-  cairo_set_source (cr, pat);
-  cairo_pattern_destroy (pat);
-  cairo_fill (cr);
-
-  os_cairo_draw_rounded_rect (cr, x, y, width, height, radius);
-
-  if (priv->orientation == GTK_ORIENTATION_VERTICAL)
-    pat = cairo_pattern_create_linear (x, y, x, height + y);
-  else
-    pat = cairo_pattern_create_linear (x, y, width + x, y);
-
-  if (priv->button_press_event && !priv->motion_notify_event)
+  if ((priv->event & OS_EVENT_BUTTON_PRESS) &&
+      !(priv->event & OS_EVENT_MOTION_NOTIFY))
     {
       if ((priv->orientation == GTK_ORIENTATION_VERTICAL && (priv->pointer_y < height / 2)) ||
           (priv->orientation == GTK_ORIENTATION_HORIZONTAL && (priv->pointer_x < width / 2)))
         {
-          state_type_up = GTK_STATE_ACTIVE;
-          cairo_pattern_add_color_stop_rgba (pat, 0.0, 0.8, 0.8, 0.8, 0.8);
-          cairo_pattern_add_color_stop_rgba (pat, 0.49, 1.0, 1.0, 1.0, 0.0);
-          cairo_pattern_add_color_stop_rgba (pat, 0.49, 0.8, 0.8, 0.8, 0.5);
-          cairo_pattern_add_color_stop_rgba (pat, 1.0, 0.8, 0.8, 0.8, 0.8);
+          pattern_add_gdk_rgba_stop (pat, 0.0, &bg_arrow_down, 0.6);
+          pattern_add_gdk_rgba_stop (pat, 0.49, &bg_arrow_down, 0.1);
+          pattern_add_gdk_rgba_stop (pat, 0.49, &bg_arrow_down, 0.1);
+          pattern_add_gdk_rgba_stop (pat, 1.0, &bg_arrow_down, 0.8);
         }
       else
         {
-          state_type_down = GTK_STATE_ACTIVE;
-          cairo_pattern_add_color_stop_rgba (pat, 0.0, 1.0, 1.0, 1.0, 0.8);
-          cairo_pattern_add_color_stop_rgba (pat, 0.49, 1.0, 1.0, 1.0, 0.0);
-          cairo_pattern_add_color_stop_rgba (pat, 0.49, 0.8, 0.8, 0.8, 0.5);
-          cairo_pattern_add_color_stop_rgba (pat, 1.0, 0.6, 0.6, 0.6, 0.8);
+          pattern_add_gdk_rgba_stop (pat, 0.0, &bg_arrow_up, 0.8);
+          pattern_add_gdk_rgba_stop (pat, 0.49, &bg_arrow_up, 0.1);
+          pattern_add_gdk_rgba_stop (pat, 0.49, &bg_arrow_down, 0.2);
+          pattern_add_gdk_rgba_stop (pat, 1.0, &bg_arrow_down, 1.0);
         }
     }
   else
     {
-      cairo_pattern_add_color_stop_rgba (pat, 0.0, 1.0, 1.0, 1.0, 0.8);
-      cairo_pattern_add_color_stop_rgba (pat, 0.49, 1.0, 1.0, 1.0, 0.0);
-      cairo_pattern_add_color_stop_rgba (pat, 0.49, 0.8, 0.8, 0.8, 0.5);
-      cairo_pattern_add_color_stop_rgba (pat, 1.0, 0.8, 0.8, 0.8, 0.8);
+      pattern_add_gdk_rgba_stop (pat, 0.0, &bg_arrow_up, 0.8);
+      pattern_add_gdk_rgba_stop (pat, 0.49, &bg_arrow_up, 0.1);
+      pattern_add_gdk_rgba_stop (pat, 0.49, &bg_arrow_down, 0.1);
+      pattern_add_gdk_rgba_stop (pat, 1.0, &bg_arrow_down, 0.8);
     }
   cairo_set_source (cr, pat);
   cairo_pattern_destroy (pat);
-  cairo_fill_preserve (cr);
 
-  cairo_set_source_rgba (cr, 0.6, 0.6, 0.6, 1.0);
-
-  if (priv->motion_notify_event)
+  if (priv->event & OS_EVENT_MOTION_NOTIFY)
     {
-      cairo_set_source_rgba (cr, 0.5, 0.5, 0.5, 0.2);
       cairo_fill_preserve (cr);
-      cairo_set_source_rgba (cr, 0.5, 0.5, 0.5, 1.0);
-      cairo_stroke (cr);
-
-      os_cairo_draw_rounded_rect (cr, x + 1, y + 1, width - 2, height - 2, radius + 1);
-      cairo_set_source_rgba (cr, 1, 1, 1, 0.5);
-      cairo_stroke (cr);
+      set_source_gdk_rgba (cr, &bg_arrow_down, 0.3);
+      cairo_fill (cr);
     }
   else
-    {
-      cairo_stroke (cr);
+    cairo_fill (cr);
 
-      os_cairo_draw_rounded_rect (cr, x + 1, y + 1, width - 2, height - 2, radius + 1);
-      cairo_set_source_rgba (cr, 1, 1, 1, 0.5);
-      cairo_stroke (cr);
-    }
+  /* 2px fat border around the thumb */
+  cairo_save (cr);
 
+  cairo_set_line_width (cr, 2.0);
+  draw_round_rect (cr, 0.5, 0.5, width - 1, height - 1, radius - 1);
+  if (!priv->detached)
+    set_source_gdk_rgba (cr, &bg_selected, 1.0);
+  else
+    set_source_gdk_rgba (cr, &bg_active, 1.0);
+  cairo_stroke (cr);
+
+  cairo_restore (cr);
+
+  /* 1px subtle shadow around the background */
+  shade_gdk_rgba (&bg, 0.2, &bg_shadow);
+
+  draw_round_rect (cr, 1, 1, width - 2, height - 2, radius);
+  set_source_gdk_rgba (cr, &bg_shadow, 0.26);
+  cairo_stroke (cr);
+
+  /* 1px frame around the background */
+  shade_gdk_rgba (&bg, 0.6, &bg_dark_line);
+
+  draw_round_rect (cr, 2, 2, width - 4, height - 4, radius - 1);
+  set_source_gdk_rgba (cr, &bg_dark_line, 0.26);
+  cairo_stroke (cr);
+
+  shade_gdk_rgba (&bg, 1.2, &bg_bright_line);
+
+  /* separators between the two steppers */
   if (priv->orientation == GTK_ORIENTATION_VERTICAL)
     {
-      cairo_move_to (cr, x + 0.5, y - 1 + height / 2);
-      cairo_line_to (cr, width - 0.5, y - 1 + height / 2);
-      cairo_set_source_rgba (cr, 0.6, 0.6, 0.6, 0.4);
+      cairo_move_to (cr, 2.5, - 1 + height / 2);
+      cairo_line_to (cr, width - 2.5, - 1 + height / 2);
+      set_source_gdk_rgba (cr, &bg_dark_line, 0.36);
       cairo_stroke (cr);
 
-      cairo_move_to (cr, x + 0.5, y + height / 2);
-      cairo_line_to (cr, width - 0.5, y + height / 2);
-      cairo_set_source_rgba (cr, 1, 1, 1, 0.5);
+      cairo_move_to (cr, 2.5, height / 2);
+      cairo_line_to (cr, width - 2.5, height / 2);
+      set_source_gdk_rgba (cr, &bg_bright_line, 0.5);
       cairo_stroke (cr);
     }
   else
     {
-      cairo_move_to (cr, x - 1 + width / 2, y + 0.5);
-      cairo_line_to (cr, x - 1 + width / 2, height - 0.5);
-      cairo_set_source_rgba (cr, 0.6, 0.6, 0.6, 0.4);
+      cairo_move_to (cr, - 1 + width / 2, 2.5);
+      cairo_line_to (cr, - 1 + width / 2, height - 2.5);
+      set_source_gdk_rgba (cr, &bg_dark_line, 0.36);
       cairo_stroke (cr);
 
-      cairo_move_to (cr, x + width / 2, y + 0.5);
-      cairo_line_to (cr, x + width / 2, height - 0.5);
-      cairo_set_source_rgba (cr, 1, 1, 1, 0.5);
+      cairo_move_to (cr, width / 2, 2.5);
+      cairo_line_to (cr, width / 2, height - 2.5);
+      set_source_gdk_rgba (cr, &bg_bright_line, 0.5);
       cairo_stroke (cr);
     }
 
+  /* arrows */
+  if (priv->orientation == GTK_ORIENTATION_VERTICAL)
+    {
+      /* direction UP. */
+      cairo_save (cr);
+      cairo_translate (cr, 8.5, 8.5);
+      cairo_rotate (cr, G_PI);  
+      draw_arrow (cr, &arrow_color, 0.5, 0, 5, 3);
+      cairo_restore (cr);
+
+      /* direction DOWN. */
+      cairo_save (cr);
+      cairo_translate (cr, 8.5, height - 8.5);
+      cairo_rotate (cr, 0);
+      draw_arrow (cr, &arrow_color, -0.5, 0, 5, 3);
+      cairo_restore (cr);
+    }
+  else
+    {
+      /* direction LEFT. */
+      cairo_save (cr);
+      cairo_translate (cr, 8.5, 8.5);
+      cairo_rotate (cr, G_PI * 0.5);  
+      draw_arrow (cr, &arrow_color, -0.5, 0, 5, 3);
+      cairo_restore (cr);
+
+      /* direction RIGHT. */
+      cairo_save (cr);
+      cairo_translate (cr, width - 8.5, 8.5);
+      cairo_rotate (cr, G_PI * 1.5);
+      draw_arrow (cr, &arrow_color, 0.5, 0, 5, 3);
+      cairo_restore (cr);
+    }
+
   cairo_restore (cr);
 
-  if (priv->orientation == GTK_ORIENTATION_VERTICAL)
-    {
-      gtk_paint_arrow (gtk_widget_get_style (widget),
-                       gtk_widget_get_window (widget),
-                       state_type_up,
-                       GTK_SHADOW_IN,
-                       NULL,
-                       widget,
-                       "arrow",
-                       GTK_ARROW_UP,
-                       FALSE,
-                       4,
-                       4,
-                       width - 8,
-                       width - 8);
-
-      gtk_paint_arrow (gtk_widget_get_style (widget),
-                       gtk_widget_get_window (widget),
-                       state_type_down,
-                       GTK_SHADOW_NONE,
-                       NULL,
-                       widget,
-                       "arrow",
-                       GTK_ARROW_DOWN,
-                       FALSE,
-                       4,
-                       height - (width - 8) - 4,
-                       width - 8,
-                       width - 8);
-    }
-  else
-    {
-      gtk_paint_arrow (gtk_widget_get_style (widget),
-                       gtk_widget_get_window (widget),
-                       state_type_up,
-                       GTK_SHADOW_IN,
-                       NULL,
-                       widget,
-                       "arrow",
-                       GTK_ARROW_LEFT,
-                       FALSE,
-                       4,
-                       4,
-                       height - 8,
-                       height - 8);
-
-      gtk_paint_arrow (gtk_widget_get_style (widget),
-                       gtk_widget_get_window (widget),
-                       state_type_down,
-                       GTK_SHADOW_NONE,
-                       NULL,
-                       widget,
-                       "arrow",
-                       GTK_ARROW_RIGHT,
-                       FALSE,
-                       width - (height - 8) - 4,
-                       4,
-                       height - 8,
-                       height - 8);
-    }
+  cairo_restore (cr);
 
+#ifndef USE_GTK3
   cairo_destroy (cr);
+#endif
 
   return FALSE;
 }
@@ -564,8 +839,8 @@ os_thumb_leave_notify_event (GtkWidget        *widget,
   /* If we exit the thumb when a button is pressed,
    * there's no need to stop the animation, it should
    * already be stopped.
-   * Stop it only if priv->button_press_event is FALSE. */
-  if (!priv->button_press_event)
+   * Stop it only if OS_EVENT_BUTTON_PRESS is not set. */
+  if (!(priv->event & OS_EVENT_BUTTON_PRESS))
     {
       if (priv->source_id != 0)
         {
@@ -573,7 +848,7 @@ os_thumb_leave_notify_event (GtkWidget        *widget,
           priv->source_id = 0;
         }
 
-      os_animation_stop (priv->animation);
+      os_animation_stop (priv->animation, NULL);
     }
 
   priv->use_tolerance = FALSE;
@@ -624,41 +899,55 @@ os_thumb_motion_notify_event (GtkWidget      *widget,
     }
 
   /* On motion, stop the fade-out. */
-  os_animation_stop (priv->animation);
-  gtk_window_set_opacity (GTK_WINDOW (widget), 1.0f);
+  os_animation_stop (priv->animation, fade_out_stop_cb);
 
   /* If you're not dragging, and you're outside
    * the tolerance pixels, enable the fade-out.
-   * priv->motion_notify_event is TRUE only on dragging,
+   * OS_EVENT_MOTION_NOTIFY is set only on dragging,
    * see code few lines below. */
-  if (!priv->motion_notify_event)
+  if (!(priv->event & OS_EVENT_MOTION_NOTIFY))
   {
     if (!priv->use_tolerance ||
-        (abs (priv->pointer_x - event->x) > TOLERANCE_PIXELS ||
-         abs (priv->pointer_y - event->y) > TOLERANCE_PIXELS))
+        (abs (priv->pointer_x - event->x) > TOLERANCE_FADE ||
+         abs (priv->pointer_y - event->y) > TOLERANCE_FADE))
       {
         priv->use_tolerance = FALSE;
         priv->source_id = g_timeout_add (TIMEOUT_FADE_OUT,
-                                         os_thumb_timeout_fade_out_cb,
+                                         timeout_fade_out_cb,
                                          thumb);
       }
   }
 
-  if (priv->button_press_event)
+  if (priv->event & OS_EVENT_BUTTON_PRESS &&
+      !(priv->event & OS_EVENT_MOTION_NOTIFY))
     {
-      if (!priv->motion_notify_event)
-        gtk_widget_queue_draw (widget);
+      if (abs (priv->pointer_x_root - event->x_root) <= TOLERANCE_MOTION &&
+          abs (priv->pointer_y_root - event->y_root) <= TOLERANCE_MOTION)
+        return FALSE;
 
-      priv->motion_notify_event = TRUE;
+      priv->event |= OS_EVENT_MOTION_NOTIFY;
+
+      gtk_widget_queue_draw (widget);
     }
 
   return FALSE;
 }
 
+
 static void
 os_thumb_screen_changed (GtkWidget *widget,
                          GdkScreen *old_screen)
 {
+#ifdef USE_GTK3
+  GdkScreen *screen;
+  GdkVisual *visual;
+
+  screen = gtk_widget_get_screen (widget);
+  visual = gdk_screen_get_rgba_visual (screen);
+
+ if (visual)
+   gtk_widget_set_visual (widget, visual);
+#else
   GdkScreen *screen;
   GdkColormap *colormap;
 
@@ -667,6 +956,7 @@ os_thumb_screen_changed (GtkWidget *widget,
 
   if (colormap)
     gtk_widget_set_colormap (widget, colormap);
+#endif
 }
 
 static gboolean
@@ -686,12 +976,7 @@ os_thumb_scroll_event (GtkWidget      *widget,
     }
 
   /* if started, stop the fade-out. */
-  os_animation_stop (priv->animation);
-  gtk_window_set_opacity (GTK_WINDOW (widget), 1.0f);
-
-  priv->source_id = g_timeout_add (TIMEOUT_FADE_OUT,
-                                   os_thumb_timeout_fade_out_cb,
-                                   thumb);
+  os_animation_stop (priv->animation, fade_out_stop_cb);
 
   return FALSE;
 }
@@ -705,12 +990,11 @@ os_thumb_unmap (GtkWidget *widget)
   thumb = OS_THUMB (widget);
   priv = thumb->priv;
 
-  priv->button_press_event = FALSE;
-  priv->motion_notify_event = FALSE;
+  priv->event = OS_EVENT_NONE;
 
   priv->use_tolerance = FALSE;
 
-  if (priv->grabbed_widget != NULL)
+  if (priv->grabbed_widget != NULL && gtk_widget_get_mapped (priv->grabbed_widget))
     gtk_grab_add (priv->grabbed_widget);
 
   GTK_WIDGET_CLASS (os_thumb_parent_class)->unmap (widget);
@@ -731,6 +1015,42 @@ os_thumb_constructor (GType                  type,
   return object;
 }
 
+static void
+os_thumb_dispose (GObject *object)
+{
+  OsThumb *thumb;
+  OsThumbPrivate *priv;
+
+  thumb = OS_THUMB (object);
+  priv = thumb->priv;
+
+  if (priv->source_id != 0)
+    {
+      g_source_remove (priv->source_id);
+      priv->source_id = 0;
+    }
+
+  if (priv->animation != NULL)
+    {
+      g_object_unref (priv->animation);
+      priv->animation = NULL;
+    }
+
+  if (priv->grabbed_widget != NULL)
+    {
+      g_object_unref (priv->grabbed_widget);
+      priv->grabbed_widget = NULL;
+    }
+
+  G_OBJECT_CLASS (os_thumb_parent_class)->dispose (object);
+}
+
+static void
+os_thumb_finalize (GObject *object)
+{
+  G_OBJECT_CLASS (os_thumb_parent_class)->finalize (object);
+}
+
 static void
 os_thumb_get_property (GObject    *object,
                        guint       prop_id,
@@ -744,14 +1064,14 @@ os_thumb_get_property (GObject    *object,
   priv = thumb->priv;
 
   switch (prop_id)
-    {
-      case PROP_ORIENTATION:
-        g_value_set_enum (value, priv->orientation);
-        break;
+  {
+    case PROP_ORIENTATION:
+      g_value_set_enum (value, priv->orientation);
+      break;
 
-      default:
-        break;
-    }
+    default:
+      break;
+  }
 }
 
 static void
@@ -767,25 +1087,20 @@ os_thumb_set_property (GObject      *object,
   priv = thumb->priv;
 
   switch (prop_id)
-    {
-      case PROP_ORIENTATION: {
+  {
+    case PROP_ORIENTATION:
+      {
         priv->orientation = g_value_get_enum (value);
         if (priv->orientation == GTK_ORIENTATION_VERTICAL)
-          {
-            gtk_window_resize (GTK_WINDOW (object), DEFAULT_THUMB_WIDTH,
-                               DEFAULT_THUMB_HEIGHT);
-          }
+          gtk_window_resize (GTK_WINDOW (object), THUMB_WIDTH, THUMB_HEIGHT);
         else
-          {
-            gtk_window_resize (GTK_WINDOW (object), DEFAULT_THUMB_HEIGHT,
-                               DEFAULT_THUMB_WIDTH);
-          }
+          gtk_window_resize (GTK_WINDOW (object), THUMB_HEIGHT, THUMB_WIDTH);
         break;
       }
 
-      default:
-        break;
-    }
+    default:
+      break;
+  }
 }
 
 /* Public functions. */
@@ -802,3 +1117,46 @@ os_thumb_new (GtkOrientation orientation)
 {
   return g_object_new (OS_TYPE_THUMB, "orientation", orientation, NULL);
 }
+
+/**
+ * os_thumb_resize:
+ * @thumb: a #OsThumb
+ * @width: width in pixels
+ * @height: height in pixels
+ *
+ * Resize the thumb.
+ **/
+void
+os_thumb_resize (OsThumb *thumb,
+                 gint     width,
+                 gint     height)
+{
+  g_return_if_fail (OS_IS_THUMB (thumb));
+
+  gtk_window_resize (GTK_WINDOW (thumb), width, height);
+}
+
+/**
+ * os_thumb_set_detached:
+ * @thumb: a #OsThumb
+ * @detached: a gboolean
+ *
+ * Sets the thumb to be detached.
+ **/
+void
+os_thumb_set_detached (OsThumb *thumb,
+                       gboolean detached)
+{
+  OsThumbPrivate *priv;
+
+  g_return_if_fail (OS_IS_THUMB (thumb));
+
+  priv = thumb->priv;
+
+  if (priv->detached != detached)
+    {
+      priv->detached = detached;
+
+      gtk_widget_queue_draw (GTK_WIDGET (thumb));
+    }
+}
